from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.commander import Command, Resource
import duplocloud.args as args

_STATUS_CODES = {
  "1": "Running",
  "3": "Pending",
  "4": "Waiting",
  "6": "Deleted",
  "7": "Failed",
  "11": "Succeeded"
}

@Resource("pod")
class DuploPod(DuploTenantResourceV2):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.paths = {
      "list": "GetPods"
    }
    self.__last_lines = {}

  @Command()
  def logs(self,
           name: args.NAME = None,
           pod: dict = None):
    """Retrieve logs for a pod."""
    if not pod:
      pod = self.find(name)
    # must be certain status for logs and a host is present
    if pod["CurrentStatus"] not in [1, 11, 7] or not pod["Host"]:
      return None
    id = pod["InstanceId"]
    data = {
      "HostName": pod["Host"],
      "DockerId": pod["Containers"][0]["DockerId"],
      "Tail": 50
    }
    response = self.duplo.post(self.endpoint("findContainerLogs"), data)
    o = response.json()
    lines = o["Data"].split("\n")
    if lines[-1] == "":
      lines.pop()
    last = self.__last_lines.get(id, 0)
    count = len(lines)
    diff = count - last
    self.__last_lines[id] = count
    if diff > 0:
      title = id
      spaces = len(title) * " "
      for line in lines[-diff:]:
        self.duplo.logger.info(f"{title}: {line}")
        title = spaces
    return None

  def name_from_body(self, body):
    return body["InstanceId"]



#!/usr/bin/env python3
import datetime
import os
import semver
from git import Repo
from packaging.version import Version
import argparse

HERE = os.path.dirname(__file__)
CWD = os.getcwd()
UNRELEASED = "## [Unreleased]"
CHANGELOG = os.path.join(CWD, 'CHANGELOG.md')
DIST = os.path.join(CWD, 'dist')
REPO = Repo(CWD)

parser = argparse.ArgumentParser(
  prog='version-bumper',
  description='A duploctl version bumper.',
)

parser.add_argument('action', type=str, help='The type of version bump to perform.', default="patch")
parser.add_argument('push', type=str, help='Push to remote?', default="false")

def get_changelog():
  with open(CHANGELOG, 'r') as f:
    return f.read()
  
def save_changelog(changelog):
  with open(CHANGELOG, 'w') as f:
    f.write(changelog)
  
def release_notes(changelog):
  inblock = False
  msg = []
  for line in changelog.splitlines():
    if line.startswith(UNRELEASED):
      inblock = True
      continue
    if inblock and line.startswith("## ["):
      break
    if inblock:
      msg.append(line.strip())
  return "\n".join(msg)

def replace_unreleased(changelog, version):
  # get todays date formatted like 2023-03-05
  date = datetime.datetime.now().strftime("%Y-%m-%d")
  new_notes = f"{UNRELEASED}\n\n## [{version}] - {date}"
  return changelog.replace(UNRELEASED, new_notes)

def latest_tag():
  latest = Version("0.0.0")
  for t in REPO.tags:
    v = Version(t.name[1:])
    if v > latest:
      latest = v
  return semver.VersionInfo.parse(str(latest))

def bump_version(action: str="patch"):
  latest = latest_tag()
  if action == "major":
    return latest.bump_major()
  if action == "minor":
    return latest.bump_minor()
  if action == "patch":
    return latest.bump_patch()
  return latest

def save_github_output(notes, version, tag):
  outputs = os.environ.get('GITHUB_OUTPUT', './.github/output')
  # append to the end of the file
  with open(outputs, 'a') as f:
    f.write(f"version={version}\n")
    f.write(f"tag={tag}\n")
  # make sure the dist folder exists
  os.makedirs(DIST, exist_ok=True)
  with open(f"{DIST}/notes.md", 'w') as f:
    f.write(notes)

def commit_changes(tag):
  msg = f"Release {tag}"
  email = os.environ.get('GITHUB_EMAIL', None)
  REPO.config_writer().set_value("user", "name", "Github Actions").release()
  REPO.config_writer().set_value("user", "email", email).release()
  print(f"Committing changes for {tag} {CHANGELOG}")
  REPO.index.add([CHANGELOG])
  REPO.index.commit(msg)
  REPO.create_tag(tag, message=msg)
  origin = REPO.remote(name='origin')
  # can't push on main branch
  # origin.push() 
  origin.push(tags=True)

def main():
  args = parser.parse_args()
  v = bump_version(args.action)
  t = f"v{v}"
  c = get_changelog()
  n = release_notes(c)
  c = replace_unreleased(c, v)
  if args.push != "true":
    v = REPO.head.reference
    t = v
  save_github_output(n, v, t)
  save_changelog(c)
  if args.push == "true":
    print(f"Pushing changes for v{v}")
    commit_changes(t)
  
if __name__ == '__main__':
  main()



from datetime import timedelta
import datetime
from duplocloud.client import DuploClient
from duplocloud.resource import DuploResource
from duplocloud.errors import DuploError
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("tenant")
class DuploTenant(DuploResource):
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
  
  @Command()
  def list(self):
    """Retrieve a list of all tenants in the Duplo system."""
    response = self.duplo.get("adminproxy/GetTenantNames")
    return response.json()

  @Command()
  def find(self, 
           name: args.NAME=None):
    """Find a tenant by name."""
    name = name or self.duplo.tenant
    try:
      return [t for t in self.list() if t["AccountName"] == name][0]
    except IndexError:
      raise DuploError(f"Tenant '{name}' not found", 404)
  
  @Command()
  def create(self, 
             body: args.BODY,
             wait: args.WAIT=False):
    """Create a new tenant."""
    self.duplo.post("admin/AddTenant", body)
    def wait_check():
      self.find(body["AccountName"])
    if wait:
      self.wait(wait_check)
    return {
      "message": f"Tenant '{body['AccountName']}' created"
    }
  
  @Command()
  def delete(self,
             name: args.NAME=None):
    """Delete a tenant."""
    name = name or self.duplo.tenant
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    self.duplo.post(f"admin/DeleteTenant/{tenant_id}", None)
    return {
      "message": f"Tenant '{name}' deleted"
    }
    
  @Command()
  def shutdown(self, 
               name: args.NAME=None, 
               schedule: args.SCHEDULE=None):
    """Expire a tenant."""
    name = name or self.duplo.tenant
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    # if the schedule not specified then set the date 5 minute from now
    if schedule is None:
      now = datetime.datetime.now() + datetime.timedelta(minutes=5)
      schedule = now.strftime("%a, %d %b %Y %H:%M:%S GMT")
    else:
      interval, unit = int(schedule[:-1]), schedule[-1]
      # Calculate the timedelta based on the specified time interval
      if unit == 'm':
        delta = timedelta(minutes=interval)
      elif unit == 'h':
        delta = timedelta(hours=interval)
      elif unit == 'd':
        delta = timedelta(days=interval)
      else:
          raise ValueError("Invalid time unit specified. Please use 'm' for minutes, 'h' for hours, or 'd' for days.")

      current_time = datetime.datetime.utcnow()
      # Calculate the future time after adding the timedelta
      future_time = current_time + delta
      # Format the future time in the desired string format
      schedule = future_time.strftime("%a, %d %b %Y %H:%M:%S GMT")

    res = self.duplo.post("adminproxy/UpdateTenantCleanupTimers", {
      "TenantId": tenant_id,
      "PauseTime": schedule
    })

    if res.status_code == 200:
      return f"Tenant '{name}' will shutdown on {schedule}"
    else:
      raise DuploError(f"Failed to expire tenant '{name}'", res.status_code)

  @Command()
  def logging(self, 
              name: args.NAME, 
              enable: args.ENABLE=True):
    """Enable or disable tenant logging."""
    name = name or self.duplo.tenant
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    # add or update the tenant in the list of enabled tenants
    response = self.duplo.get("admin/GetLoggingEnabledTenants")
    log_tenants = response.json()
    for t in log_tenants:
      if t["TenantId"] == tenant_id:
        t["Enabled"] = enable
        break
    else:
      log_tenants.append({"TenantId": tenant_id, "Enabled": enable})
    print(log_tenants)
    # update the entire list
    res = self.duplo.post("admin/UpdateLoggingEnabledTenants", log_tenants)
    if res.status_code == 200:
      return f"Tenant '{name}' logging {enable}"
    else:
      raise DuploError(f"Failed to {'enable' if enable else 'disable'} tenant '{name}'", res.status_code)

  @Command()
  def billing(self,
              name: args.NAME=None):
    """Spend
    
    Get the spend for the tenant. 
    """
    tenant = self.find(name or self.duplo.tenant)
    tenant_id = tenant["TenantId"]
    response = self.duplo.get(f"v3/billing/subscriptions/{tenant_id}/aws/billing")
    return response.json()
  
  @Command()
  def config(self,
             name: args.NAME=None,
             setvar: args.SETVAR=[],
             deletevar: args.DELETEVAR=[]):
    """Add a setting to the tenant."""
    updates = []
    creates = []
    changes = []
    name = name or self.duplo.tenant
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    curr_settings = tenant.get("Metadata", [])
    curr_keys = [s["Key"] for s in curr_settings]
    # flatten k/v pair while dedupe and remove deleted keys
    new_settings = {s[0]: s[1] for s in setvar if s[0] not in deletevar}
    # only update if the value is different and create if the key is not present
    for k, v in new_settings.items():
      s = {"Key": k, "Value": v}
      for c in curr_settings:
        if c["Key"] == k:
          if c["Value"] != v:
            updates.append(s)
          break
      else:
        creates.append(s)
    # create, update, and delete the settings
    for s in creates:
      response = self.duplo.post(f"v3/admin/tenant/{tenant_id}/metadata", s)
      change = response.json()
      change["Operation"] = "create"
      changes.append(change)
    for s in updates:
      response = self.duplo.put(f"v3/admin/tenant/{tenant_id}/metadata/{s['Key']}", s)
      change = response.json()
      change["Operation"] = "update"
      changes.append(change)
    for k in deletevar:
      if k in curr_keys:
        self.duplo.delete(f"v3/admin/tenant/{tenant_id}/metadata/{k}")
        change = {"Key": k, "Operation": "delete"}
        changes.append(change)
    return {
      "message": f"Successfully updated settings for tenant '{name}'",
      "changes": changes
    }
    
  @Command()
  def host_images(self,
                  name: args.NAME = None):
    """Get the list of host images."""
    tenant = self.find(name or self.duplo.tenant)
    tenant_id = tenant["TenantId"]
    response = self.duplo.get(f"v3/subscriptions/{tenant_id}/nativeHostImages")
    return response.json()

  @Command()
  def faults(self,
             name: args.NAME = None):
    """Get the list of faults."""
    tenant = self.find(name or self.duplo.tenant)
    tenant_id = tenant["TenantId"]
    response = self.duplo.get(f"subscriptions/{tenant_id}/GetFaultsByTenant")
    return response.json()



from duplocloud.client import DuploClient
from duplocloud.resource import DuploResource
from duplocloud.commander import Command, Resource

@Resource("system")
class DuploSystem(DuploResource):
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
  
  @Command()
  def info(self):
    """Retrieve all of the system information."""
    return self.duplo.get("v3/features/system").json()
  
  @Command()
  def billing(self):
    """Account
    
    Get the account spend for the portal. 
    """
    response = self.duplo.get("v3/billing/admin/aws/billing")
    return response.json()




import requests
import jmespath
import os
import yaml
import json
import jsonpatch
import logging
from datetime import datetime, timezone, timedelta
from urllib.parse import urlparse
from cachetools import cachedmethod, TTLCache
from pathlib import Path
from .commander import load_resource,load_format
from .errors import DuploError, DuploExpiredCache
from .server import TokenServer
from . import args
from .commander import Command, get_parser

class DuploClient():
  """Duplo Client

  This is the main Duplo client class. It is used to connect to Duplo and
  to retrieve resources from Duplo. All services have a reference to this
  when they are created.

  Example: Using injected client to load a service.
      ```python
      from duplocloud.client import DuploClient  
      from duplocloud.resource import DuploResource  
      from duplocloud.errors import DuploError  

      class DuploSomeService(DuploResource):
        def __init__(self, duplo: DuploClient):
          super().__init__(duplo)
          self.tenent_svc = duplo.service('tenant')
      ```
  """
  @Command()
  def __init__(self, 
               host: args.HOST=None,
               token: args.TOKEN=None,
               tenant: args.TENANT=None,
               home_dir: args.HOME_DIR = None,
               config_file: args.CONFIG = None,
               cache_dir: args.CACHE_DIR = None,
               version: args.VERSION=False,
               interactive: args.INTERACTIVE=False,
               ctx: args.CONTEXT=None,
               nocache: args.NOCACHE=False,
               browser: args.BROWSER=None,
               isadmin: args.ISADMIN=False,
               query: args.QUERY=None,
               output: args.OUTPUT="json",
               loglevel: args.LOGLEVEL="INFO"):
    """DuploClient
    
    Creates an instance of a duplocloud client configured for a certain portal. All of the specific configuration is done in the DuploConfig class.
    """
    # forces the given context to be used
    if ctx: 
      host = None
      token = None
    # ignore the given token with interactive mode
    if token and interactive: 
      token = None

    user_home = Path.home()
    self.home_dir = home_dir or f"{user_home}/.duplo"
    self.config_file = config_file or f"{self.home_dir}/config"
    self.cache_dir = cache_dir or f"{self.home_dir}/cache"
    self.__config = None
    self.__context = ctx
    self.__host = self.__sanitize_host(host)
    self.__token = token.strip() if token else token
    self.__tenant = tenant.strip().lower() if tenant else tenant
    self.version = version
    self.interactive = interactive
    self.nocache = nocache
    self.browser = browser
    self.isadmin = isadmin
    self.query = query.strip() if query else query
    self.output = output.strip()
    self.timeout = 60
    self.__ttl_cache = TTLCache(maxsize=128, ttl=10)
    self.loglevel = loglevel
    self.logger = self.logger_for()

  @staticmethod
  def from_env():
    """DuploClient from Environment

    Create a DuploClient from environment variables.

    Returns:
      An instance of DuploConfig.
    """
    p = get_parser(DuploClient.__init__)
    env, xtra = p.parse_known_args()
    duplo = DuploClient(**vars(env))
    return duplo, xtra
  
  @staticmethod
  def from_args(*args):
    """DuploClient from Environment

    Create a DuploClient from environment variables.

    Returns:
      An instance of DuploConfig.
    """
    p = get_parser(DuploClient.__init__)
    env = p.parse_args(args)
    duplo = DuploClient(**vars(env))
    return duplo
  
  @staticmethod
  def from_creds(host: str, token: str, tenant: str):
    """Create a DuploClient from credentials.
    
    Args:
      host: The host of the Duplo instance.
      token: The token to use for authentication.
      tenant: The tenant to use.
    Returns:
      The DuploClient.
    """
    return DuploClient(host=host, token=token, tenant=tenant)

  @property
  def settings(self):
    """Get Config

    Get the Duplo config as a dict. This is accessed as a lazy loaded property.

    Returns:
      The config as a dict.
    """
    if self.__config is None:
      if not os.path.exists(self.config_file):
        raise DuploError("Duplo config not found", 500)
      with open(self.config_file, "r") as f:
        self.__config = yaml.safe_load(f)
    return self.__config
  
  @property
  def context(self):
    """Get Config Context
    
    Get the current context from the Duplo config. This is accessed as a lazy loaded property. 

    Returns:
      The context as a dict.
    """
    s = self.settings
    ctx = self.__context if self.__context else s.get("current-context", None)
    if ctx is None:
      raise DuploError(
        "Duplo context not set, please set 'current-context' to a portals name in your config", 500)
    try:
      return [c for c in s["contexts"] if c["name"] == ctx][0]
    except IndexError:
      raise DuploError(f"Portal '{ctx}' not found in config", 500)
    
  @property
  def host(self):
    """Get Host
    
    Get the host from the Duplo config. This is accessed as a lazy loaded property. 
    If the host is some kind of falsey value, it will attempt to use the context.

    Returns:
      The host as a string.
    """
    if not self.__host:
      self.use_context()
    return self.__host
  
  @property 
  def token(self):
    """Get Token
    
    Returns the configured token. If interactive mode is enabled, an attempt will be made to get the token interactively. Ultimately, the token is required and if it is not set, an error will be raised.
    This is accessed as a lazy loaded property.

    Returns:
      The token as a string.
    """
    if not self.host:
      raise DuploError("Host for Duplo portal is required", 500)
    if not self.__token and self.interactive:
      self.__token = self.interactive_token()
    if not self.__token:
      raise DuploError("Token for Duplo portal is required", 500)
    return self.__token
  
  @property
  def tenant(self):
    """Get Tenant
    
    Get the tenant from the Duplo config. This is accessed as a lazy loaded property. 
    If the tenant is some kind of falsey value, it will attempt to use the context.

    Returns:
      The tenant as a string.
    """
    if not self.host:
      raise DuploError("Host for Duplo portal is required", 500)
    return self.__tenant
  
  @tenant.setter
  def tenant(self, value: str):
    """Set Tenant
    
    Set the tenant for this Duplo client. This will override the tenant in the config.

    Args:
      value: The tenant to set.
    """
    self.__tenant = value
  
  def __str__(self) -> str:
     return f"""
Client for Duplo at {self.host}
"""
  
  def __call__(self, resource: str, *args):
    """Run a service command.
    
    Args:
      resource: The name of the resource.
      command: The command to run.
      args: The arguments to the command.
    Returns:
      The result of the command.
    """
    r = self.load(resource)
    d = r(*args)
    if d is None:
      return None
    d = self.filter(d)
    return self.format(d)
  
  def logger_for(self, name: str=None) -> logging.Logger:
    """Create a Default Logger
    
    Create a default logger for the given name. This will create a logger with the name 'duplo' and add a console output handler.

    Args:
      name: The name of the logger.
    Returns:
      The logger.
    """
    n = "duplo"
    if name:
      n += f".{name}"
    logger = logging.getLogger(name)
    lvl = logging.getLevelName(self.loglevel)
    logger.setLevel(lvl)
    formatter = logging.Formatter("%(levelname)s %(message)s")
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    if (logger.hasHandlers()):
      logger.handlers.clear()
    logger.addHandler(handler)
    return logger

  @cachedmethod(lambda self: self.__ttl_cache)
  def get(self, path: str):
    """Get a Duplo resource.

    This request is cached for 60 seconds.
    
    Args:
      path: The path to the resource.
    Returns:
      The resource as a JSON object.
    """
    try:
      response = requests.get(
        url = f"{self.host}/{path}",
        headers = self.__headers(),
        timeout = self.timeout
      )
    except requests.exceptions.Timeout as e:
      raise DuploError("Timeout connecting to Duplo", 500) from e
    except requests.exceptions.ConnectionError as e:
      raise DuploError("A conntection error occured with Duplo", 500) from e
    except requests.exceptions.RequestException as e:
      raise DuploError("Error connecting to Duplo with a request exception", 500) from e
    return self.__validate_response(response)
  
  def post(self, path: str, data: dict={}):
    """Post data to a Duplo resource.
    
    Args:
      path: The path to the resource.
      data: The data to post.
    Returns:
      The response as a JSON object.
    """
    response = requests.post(
      url = f"{self.host}/{path}",
      headers = self.__headers(),
      timeout = self.timeout,
      json = data
    )
    return self.__validate_response(response)
  
  def put(self, path: str, data: dict={}):
    """Put data to a Duplo resource.
    
    Args:
      path: The path to the resource.
      data: The data to post.
    Returns:
      The response as a JSON object.
    """
    response = requests.put(
      url = f"{self.host}/{path}",
      headers = self.__headers(),
      timeout = self.timeout,
      json = data
    )
    return self.__validate_response(response)
  
  def delete(self, path: str):
    """Delete a Duplo resource.
    
    Args:
      path: The path to the resource.
    Returns:
      The response as a JSON object.
    """
    response = requests.delete(
      url = f"{self.host}/{path}",
      headers = self.__headers(),
      timeout = self.timeout
    )
    return self.__validate_response(response)
  
  def jsonpatch(self, data, patches):
    """Json Patch
    
    Apply a json patch to a resource.

    Args:
      patches: The patches to apply.
    Returns:
      The patched resource as a JSON object.
    """
    try:
      return jsonpatch.apply_patch(data, patches)
    except jsonpatch.JsonPatchTestFailed as e:
      raise DuploError("JsonPatch test failed", 500) from e
    except jsonpatch.JsonPatchConflict as e:
      raise DuploError(f"JsonPatch conflict:\n {e}", 500)

  def filter(self, data: dict):
    """Query data

    Uses the jmespath library to query data.
    Set the query to use on the property. 
    
    Args:
      data: The data to query.
    Returns:
      The queried data.
    """
    if not self.query:
      return data
    try:
      return jmespath.search(self.query, data)
    except jmespath.exceptions.ParseError as e:
      raise DuploError("Invalid jmespath query", 500) from e
    except jmespath.exceptions.JMESPathTypeError as e:
      raise DuploError("Invalid jmespath query", 500) from e
    
  def load(self, resource: str):
    """Load Service
      
    Load a resource class from the entry points.

    Args:
      name: The name of the service.
    Returns:
      The instantiated service with a reference to this client.
    """
    svc = load_resource(resource)
    return svc(self)
  
  def format(self, data: dict) -> str:
    """Format data.
    
    Args:
      data: The data to format.
    Returns:
      The data as a string.
    """
    fmt = load_format(self.output)
    return fmt(data)
  
  def use_context(self, name: str = None) -> None:
    """Use Context
    
    Use the specified context from the Duplo config.

    Args:
      name: The name of the context to use.
    """

    # Get the right context
    if name:
      self.__context = name
    ctx = self.context

    # set the context into this config
    self.__host = self.__sanitize_host(ctx.get("host", None))
    self.__token = ctx.get("token", None)
    self.__tenant = ctx.get("tenant", self.__tenant)
    self.interactive = ctx.get("interactive", False)
    self.isadmin = ctx.get("admin", False)
    self.nocache = ctx.get("nocache", False)
    # only tenant can be overridden by the args/env
    
  def get_cached_item(self, key: str) -> dict:
    """Get Cached Item
    
    Get a cached item from the cache directory. The files are all json. 
    This checks if the file exists and raises an expired cache if it does not. 
    Finally the file is read and returned as a JSON object. If anything goes wrong, an expired cache is raised because it's easy enough to rebuild it than to try and fix it.

    Args:
      name: The name of the item to get.

    Returns:
      The json content parsed as a dict.
    """
    fn = f"{self.cache_dir}/{key}.json"
    if not os.path.exists(fn):
      raise DuploExpiredCache(key)
    try:
      with open(fn, "r") as f:
        return json.load(f)
    except json.JSONDecodeError:
      raise DuploExpiredCache(key)
    
  def set_cached_item(self, key: str, data: dict) -> None:
    """Set Cached Item
    
    Set a cached item in the cache directory. The files are all json. 
    This writes the data to the file as a JSON object.

    Args:
      key: The key of the item to set.
      data: The data to set.
    """
    if not os.path.exists(self.cache_dir):
      os.makedirs(self.cache_dir)
    fn = f"{self.cache_dir}/{key}.json"
    with open(fn, "w") as f:
      json.dump(data, f)

  def interactive_token(self) -> str:
    """Interactive Token
    
    Performs an interactive login for the configured host. The cache will be checked for a token and if it is expired, it will perform an interactive login and cache the token. The cache may be disabled by setting the nocache flag. 
    
    Returns:
      The token as a string.
    """
    t = None
    k = self.cache_key_for("duplo-creds")
    try:
      if self.nocache:
        t = self.request_token()
      else:
        t = self.cached_token(k)
    except DuploExpiredCache:
      t = self.request_token()
      c = self.__token_cache(t)
      self.set_cached_item(k, c)
    return t
  
  def cached_token(self, key: str) -> str:
    """Cached Token
    
    Get a cached token from the cache directory. This checks if the file exists and raises a 404 if it does not. 
    Finally the file is read and returned as a JSON object. 
    The Expiration key looks like: "2024-01-12T18:51:48Z" in the popular iso8601 format.

    Args:
      host: The host to get the token for.

    Returns:
      The token as a string.
    """
    c = self.get_cached_item(key)
    if (exp := c.get("Expiration", None)) and (t := c.get("DuploToken", None)):
      if not self.expired(exp):
        return t
    raise DuploExpiredCache(key)
  
  def request_token(self) -> str:
    """Request Token from Browser
    
    Perform an interactive login to the specified host. Opens a temporary web browser to the login page and starts a local server to receive the token. When the user authorizes the request in the browser, the token is received and the server is shutdown.

    Returns:
      The token as a string.
    """
    isadmin = "true" if self.isadmin else "false"
    path = "app/user/verify-token"
    with TokenServer(self.host) as server:
      try:
        page = f"{path}?localAppName=duploctl&localPort={server.server_port}&isAdmin={isadmin}"
        server.open_callback(page, self.browser)
        return server.serve_token()
      except KeyboardInterrupt:
        server.shutdown()

  def cache_key_for(self, name: str) -> str:
    """Cache Key For
    
    Get the cache key for the given name. This is a simple string concatenation of the host and the name.

    Args:
      name: The name to get the cache key for.

    Returns:
      The cache key as a string.
    """
    h = self.host.split("://")[1].replace("/", "")
    parts = [h]
    if self.isadmin:
      parts.append("admin")
    parts.append(name)
    return ",".join(parts)
  
  def expiration(self, hours: int = 6) -> str:
    """Expiration
    
    Get the expiration time for the given number of hours. This is a simple calculation of the current time plus the number of hours.

    Args:
      hours: The number of seconds to add to the current time.

    Returns:
      The expiration time as a string.
    """

    return (datetime.now(timezone.utc) + timedelta(hours=hours)).strftime('%Y-%m-%dT%H:%M:%S+00:00')
  
  def expired(self, exp: str = None) -> bool:
    """Expired
    
    Check if the given expiration time is expired. This is a simple comparison of the current time and the expiration time.

    Args:
      exp: The expiration time to check.

    Returns:
      True if the expiration time is in the past, False otherwise.
    """
    if exp is None:
      return True
    return datetime.now(timezone.utc) > datetime.fromisoformat(exp)
  
  def build_command(self, *args) -> list[str]:
    """Context Args
    
    Build a comamnd using the current context. 

    Returns:
      The context args as a dict.
    """
    cmd = list(args)
    cmd.append("--host")
    cmd.append(self.host)
    if self.isadmin:
      cmd.append("--admin")
    if self.interactive:
      cmd.append("--interactive")
      if self.nocache:
        cmd.append("--nocache")
      if self.browser:
        cmd.append("--browser")
        cmd.append(self.browser)
    else:
      cmd.append("--token")
      cmd.append(self.token)
    return cmd
  
  def disable_get_cache(self) -> None:
    """Disable Get Cache
    
    Disable the get cache for this client. This is useful for testing.
    Disable by setting the __ttl_cache to None.
    """
    self.__ttl_cache = None

  def __token_cache(self, token, otp=False) -> dict:
    return {
      "Version": "v1",
      "DuploToken": token,
      "Expiration": self.expiration(),
      "NeedOTP": otp
    }
  
  def __headers(self) -> dict:
    t = self.token
    return {
      'Content-Type': 'application/json',
      'Authorization': f"Bearer {t}"
    }
  
  def __validate_response(self, response: requests.Response) -> requests.Response:
    """Validate a response from Duplo.
    
    Args:
      response: The response to validate.
    Raises:
      DuploError: If the response was not 200. 
    Returns:
      The response as a JSON object.
    """
    # contentType = response.headers.get('content-type', 'application/json').split(';')[0]
    if 200 <= response.status_code < 300:
      return response
    
    if response.status_code == 404:
      raise DuploError("Resource not found", response.status_code)
    
    if response.status_code == 401:
      raise DuploError(response.text, response.status_code)
    
    if response.status_code == 403:
      raise DuploError("Unauthorized", response.status_code)  
    
    if response.status_code == 400:
      raise DuploError(response.text, response.status_code)

    raise DuploError("Duplo responded with an error", response.status_code)
    
  def __sanitize_host(self, host: str) -> str:
    """Sanitize Host
    
    Sanitize the host using urlparse. This will ensure that the host is a valid URL and that it is using HTTPS.

    Args:
      host: The host to sanitize.
    Returns:
      The sanitized host with scheme.
    """
    if not host:
      return None
    url = urlparse(host)
    return f"https://{url.netloc}"
  
  




def tostring(obj):
  """Converts a python object to a string.

  The default formatter with no fancy pants tranforming.

  Args:
    obj: The python object to convert.

  Returns:
    A string representing the object.
  """
  return str(obj)

def tojson(obj):
  """Converts a python object to a JSON string.

  Args:
    obj: The python object to convert.

  Returns:
    A JSON string representing the object.
  """
  import json
  return json.dumps(obj)

def toyaml(obj):
  """Converts a python object to a YAML string.

  Args:
    obj: The python object to convert.

  Returns:
    A YAML string representing the object.
  """
  import yaml
  return yaml.dump(obj)

def toenv(obj):
  """Converts a python object to a .env file format
  
  Args:
    obj: The python object to convert.
  
  Returns:
    A .env file format string representing the object.
  """
  envs = [f"{k}={v}" for k, v in obj.items()]
  return "\n".join(envs)

def tocsv(obj):
  """Converts a python object to a CSV string.

  Args:
    obj: The python object to convert.

  Returns:
    A CSV string representing the object.
  """
  import csv
  from io import StringIO
  if not isinstance(obj, list):
    obj = [obj]
  headers = obj[0].keys()
  output = StringIO()
  writer = csv.writer(output)
  writer.writerow(headers)
  for row in obj:
    writer.writerow(row.values())
  return output.getvalue()



#!/usr/bin/env python3
import os
import requests
import toml
import re
from importlib.metadata import version

def make_resource(name):
  v = version(name)
  url = f"https://pypi.org/pypi/{name}/{v}/json"
  response = requests.get(url)
  data = response.json()
  sdist = list(filter(lambda x: x['packagetype'] == 'sdist', data['urls']))[0]
  return f"""
    resource "{name}" do
      url "{sdist['url']}"
      sha256 "{sdist['digests']['sha256']}"
    end
  """

# get started
f = open('pyproject.toml')
v = os.sys.argv[1].replace('v', '')

# Build the homebrew pip resources from the project dependencies
data = toml.load(f)
pattern = '|'.join(map(re.escape, ['>=', '<=', '==', '!=']))
resources = []
for dep in data['project']['dependencies']:
  name, _ = re.split(pattern, dep, 1)
  res = make_resource(name)
  resources.append(res)

# get the checksums from the github release
repo_url = data['project']['urls']['Repository']
description = data['project']['description']
url = f"{repo_url}/releases/download/v{v}/checksums.txt"
response = requests.get(url)
checksums = response.text.splitlines()
linux_sha = None 
macos_sha = None
pip_sha = None
for line in checksums:
  l = line.split(" ")
  if 'darwin' in line:
    macos_sha = l[0]
  elif 'linux' in line:
    linux_sha = l[0]
  elif 'duplocloud-client' in line:
    pip_sha = l[0]

# finally build and print the formula from template
with open('scripts/formula.tpl.rb', 'r') as tpl_file:
  tpl = tpl_file.read()
  formula = tpl.format(
    repo_url=repo_url,
    description=description,
    version=v, 
    linux_sha=linux_sha, 
    macos_sha=macos_sha,
    pip_sha=pip_sha,
    resources="".join(resources)
  )
  os.makedirs('dist', exist_ok=True)
  with open('dist/duploctl.rb', 'w') as formula_file:
    formula_file.write(formula)
    print(formula)



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError, DuploFailedResource
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("hosts")
class DuploHosts(DuploTenantResourceV2):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.paths = {
      "list": "GetNativeHosts"
    }
  
  @Command()
  def create(self,
             body: args.BODY,
             wait: args.WAIT=False):
    """Create a host."""
    def wait_check():
      name = self.name_from_body(body)
      h = self.find(name)
      if h["Status"] != "running":
        if h["Status"] != "pending":
          raise DuploFailedResource(f"Host '{name}' failed to create.")
        raise DuploError(f"Host '{name}' not ready", 404)
    # let's get started
    if body.get("ImageId", None) is None:
      body["ImageId"] = self.discover_image(body.get("AgentPlatform", 0))
    res = self.duplo.post(self.endpoint("CreateNativeHost"), body)
    if wait:
      self.wait(wait_check)
    return {
      "message": f"Successfully created host '{body['FriendlyName']}'",
      "id": res.json()
    }
  
  @Command()
  def delete(self,
             name: args.NAME,
             wait: args.WAIT=False):
    """Delete a host."""
    host = self.find(name)
    inst_id = host["InstanceId"]
    res = self.duplo.post(self.endpoint(f"TerminateNativeHost/{inst_id}"), host)
    def wait_check():
      h = None 
      try:
        h = self.find(name)
      except DuploError as e:
        if e.code == 404:
          return None # if 404 then it's gone so finish waiting
        else:
          raise DuploFailedResource(f"Host '{name}' failed to delete.")
      if h["Status"] == "shutting-down" or h["Status"] == "running":
        raise DuploError(f"Host '{name}' not terminated", 404)
    if wait:
      self.wait(wait_check, 500)
    return {
      "message": f"Successfully deleted host '{name}'",
      "data": res.json()
    }
    
  def name_from_body(self, body):
    prefix = f"duploservices-{self.duplo.tenant}"
    name =  body["FriendlyName"]
    if not name.startswith(prefix):
      name = f"{prefix}-{name}"
    return name
  
  def discover_image(self, agent, arch="amd64"):
    imgs = self.tenant_svc.host_images(self.duplo.tenant)
    try:
      img = [i for i in imgs if i["Agent"] == agent and i["Arch"] == arch][0]
      return img.get("ImageId")
    except IndexError:
      raise DuploError(f"Image for agent '{agent}' not found", 404)



import pytest
import random
import yaml
import pathlib
from duplocloud.client import DuploClient

def pytest_addoption(parser):
  parser.addoption("--e2e", action="store_true", default=False, help="run e2e tests")
  parser.addoption("--infra", action="store", default=None, help="Choose existing infra")
  parser.addoption("--tenant", action="store", default=None, help="Choose existing tenant to use")

@pytest.fixture(scope='session', autouse=True)
def infra_name(pytestconfig) -> str:
  existing = pytestconfig.getoption("infra")
  if existing:
    return existing
  inc = random.randint(1, 100)
  return f"duploctl{inc}"

@pytest.fixture(scope='session', autouse=True)
def e2e(pytestconfig) -> bool:
  return pytestconfig.getoption("e2e")

@pytest.fixture(scope='session', autouse=True)
def duplo(infra_name: str, e2e: bool) -> DuploClient:
  d, _ = DuploClient.from_env()
  d.disable_get_cache()
  if e2e:
    d.tenant = infra_name
  return d

@pytest.fixture(scope="session", autouse=True)
def cleanup(request):
  """Cleanup a testing directory once we are finished."""
  def kill_infra():
    print(f"Totally cleaning up dude.")
  request.addfinalizer(kill_infra)

@pytest.fixture
def test_data(request) -> tuple[str, dict]:
  """Fixture to load test data from a yaml file.
  
  Splits like this: kind::file
  example with a data file named big_host with host data would be: host::big_host
  """
  test_id = request.param.split("::")
  kind = test_id[0]
  file = test_id[-1]
  print(f"Loading test data for {kind} from {file}")
  data = get_test_data(file)
  return (kind, data)

def get_test_data(name) -> dict:
  # get the directory this file is in
  dir = pathlib.Path(__file__).parent.resolve()
  f = f"{dir}/data/{name}.yaml"
  with open(f, 'r') as stream:
    return yaml.safe_load(stream)
